# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# Copyright (C) 2006 Alec Thomas <alec@swapoff.org>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Alec Thomas <alec@swapoff.org>

import time

from trac.core import *
from trac.config import ExtensionOption, IntOption
from trac.web.api import IRequestFilter, IRequestHandler
from tracspamfilter.api import IFilterStrategy, IRejectHandler, RejectContent


class ICaptchaMethod(Interface):
    """ A CAPTCHA implementation. """

    def generate_captcha(req):
        """Return a tuple of `(result, html)`, where `result` is the expected
        response and `html` is a HTML fragment for displaying the CAPTCHA
        challenge."""

    def verify_captcha(req):
        """Checks if captcha is valid (only in case generate_captcha returns
        None for challenge parameter, otherwise check directly."""

    def is_usable(req):
        """Check if captcha can be used for the request."""

class CaptchaSystem(Component):
    implements(IRequestHandler, IRejectHandler, IFilterStrategy, IRequestFilter)

    handlers = ExtensionPoint(IRequestHandler)

    captcha = ExtensionOption('spam-filter', 'captcha', ICaptchaMethod,
                              'ExpressionCaptcha',
        """ CAPTCHA method to use for verifying humans. """)

    karma_points = IntOption('spam-filter', 'captcha_karma', 10,
        """By how many points a successful CAPTCHA response increases the
        overall score.""")

    karma_lifetime = IntOption('spam-filter', 'captcha_karma_lifetime', 86400,
        """Time in seconds that a successful CAPTCHA response increases
        karma.""")

    captcha_lifetime = IntOption('spam-filter', 'captcha_lifetime', 120,
        """Time in seconds before CAPTCHA is removed.""")

    captcha_cleantime = IntOption('spam-filter', 'captcha_lifetime', 3600,
        """Time in seconds before database cleanup is called.""")

    # IFilterStrategy methods

    def test(self, req, author, content, ip):
        if not self._expired(req):
            return self.karma_points, 'Human verified via CAPTCHA'

    def train(self, req, author, content, ip, spam=True):
        pass

    # IRejectHandler methods

    def reject_content(self, req, message):
        self._cleanup()
        if self._expired(req):
            req.session['captcha_reject_time'] = int(time.time())
            req.session['captcha_reject_reason'] = message
            req.session['captcha_redirect'] = req.href(req.path_info)
            for key, value in req.args.iteritems():
                req.session['captcha_arg_%s' % (key)] = value
            req.redirect(req.href.captcha())
        else:
            raise RejectContent(message)

    # IRequestHandler methods

    def match_request(self, req):
        return req.path_info == '/captcha'

    def process_request(self, req):
        data = {}
        exp = req.session.get('captcha_expected')
        if req.method == 'POST':
            if req.args.get('captcha_response') == exp:
                data['error'] = 'CAPTCHA failed to handle original request'
            else:
                data['error'] = 'CAPTCHA verification failed'
        else:
            data['error'] = req.session.get('captcha_reject_reason')
        # cleanup olf values
        if exp != None:
            del req.session['captcha_expected']
        # generate new captcha
        result, html = self.captcha.generate_captcha(req)
        data['challenge'] = html
        if result != None:
            data['defaultform'] = 1
            req.session['captcha_expected'] = result
        req.session.save()
        return 'verify_captcha.html', data, None

    def pre_process_request(self, req, handler):
        if req.path_info == '/captcha' and req.method == 'POST':
            valid = False
            exp = req.session.get('captcha_expected')
            if exp == None:
                valid = self.captcha.verify_captcha(req)
            elif req.args['captcha_response'] == exp:
                valid = True

            if valid:
                req.environ['PATH_INFO'] = req.session.get('captcha_redirect', req.href())
                if ('SCRIPT_NAME' in req.environ and len(req.environ['SCRIPT_NAME']) > 1):
                    req.environ['PATH_INFO'] = req.environ['PATH_INFO'].replace(req.environ['SCRIPT_NAME'], '')
                req.environ['PATH_INFO'] = req.environ['PATH_INFO'].encode('utf-8')
                if 'captcha_redirect' in req.session:
                    del req.session['captcha_redirect']
                if 'captcha_reject_reason' in req.session:
                    del req.session['captcha_reject_reason']
                if 'captcha_reject_time' in req.session:
                    del req.session['captcha_reject_time']
                keys = req.session.keys()
                for key in keys:
                    if key.startswith('captcha_arg_'):
                        arg = key[12:]
                        req.args[arg] = req.session[key]
                        del req.session[key]
                try:
                    for newhandler in self.handlers:
                        try:
                            if newhandler.match_request(req):
                                keys = req.session.keys()
                                for key in keys:
                                    if key.startswith('captcha_'):
                                        self.log.warning('Del %s', key)
                                        del req.session[key]
                                handler = newhandler
                                break
                        except Exception, e:
                            self.log.debug('Exception when parsing handlers: (%s)', e)
                except TracError, e:
                    return (handler)
                req.session['captcha_verified'] = int(time.time())
                req.session.save()
        return (handler)

    def post_process_request(self, req, template, content_type):
        return (template, content_type)

    def post_process_request(self, req, template, data, content_type):
        return (template, data, content_type)

    # Internal methods

    def _expired(self, req):
        return int(req.session.get('captcha_verified', 0)) + \
               self.karma_lifetime < time.time()

    # remove old entries from database
    def _cleanup(self):
        last = 0
        db = self.env.get_read_db()
        cursor = db.cursor()
        cursor.execute("SELECT value FROM system "
                       "WHERE name='spamfilter_lastclean'")
        try:
            last = int(cursor.fetchone()[0])
        except:
            pass
        tim = int(time.time())
        if last+self.captcha_cleantime < tim:
            #self.log.warn('Cleanup captcha')
            t = tim - self.captcha_lifetime
            tc = tim - self.karma_lifetime
            db = self.env.get_db_cnx()
            if last == 0:
                cursor.execute("INSERT INTO system VALUES "
                               "('spamfilter_lastclean',%s)", (tim,))
            else:
                cursor.execute("UPDATE system SET value=%s WHERE "
                               "name='spamfilter_lastclean'", (tim,))
            db.commit()
            cursor = db.cursor() 
            cursor.execute(("DELETE FROM session_attribute WHERE name LIKE " \
                + "'captcha%%' AND (name != 'captcha_verified' OR %s < %%s) " \
                + "AND ("\
                + "sid IN (SELECT sid FROM session_attribute WHERE name " \
                + "= 'captcha_reject_time' AND %s < %%s) "\
                + "OR sid NOT IN (SELECT sid FROM session_attribute WHERE " \
                + "name = 'captcha_reject_time'))") %(db.cast('value', 'int'), \
                db.cast('value', 'int')), (tc,t))
            db.commit()
    
