# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Edgewall Software
# Copyright (C) 2011 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

from email.Utils import parseaddr
from pkg_resources import get_distribution
from xmlrpclib import ServerProxy, Transport, Error
import httplib

from trac import __version__ as TRAC_VERSION
from trac.config import IntOption, Option, ListOption
from trac.core import *
from tracspamfilter.api import IFilterStrategy
from trac.mimeview.api import is_binary

class TimeoutHTTPConnection(httplib.HTTPConnection):
   def connect(self):
       httplib.HTTPConnection.connect(self)
       self.sock.settimeout(self.timeout)

class TimeoutHTTP(httplib.HTTP):
   _connection_class = TimeoutHTTPConnection
   def set_timeout(self, timeout):
       self._conn.timeout = timeout

class TimeoutTransport(Transport):
    def __init__(self, timeout=3, *l, **kw):
        Transport.__init__(self,*l,**kw)
        self.timeout=timeout
    def make_connection(self, host):
        conn = TimeoutHTTP(host)
        conn.set_timeout(self.timeout)
        return conn

class TimeoutServerProxy(ServerProxy):
    def __init__(self,uri,timeout=3,*l,**kw):
        kw['transport']=TimeoutTransport(timeout=timeout, use_datetime=kw.get('use_datetime',0))
        ServerProxy.__init__(self,uri,*l,**kw)

class BlogSpamFilterStrategy(Component):
    """Spam filter using the BlogSpam service (http://blogspam.net/).
    """
    implements(IFilterStrategy)
    
    karma_points = IntOption('spam-filter', 'blogspam_karma', '5',
        """By how many points an BlogSpam reject impacts the overall karma of
        a submission.""")

    api_url = Option('spam-filter', 'blogspam_api_url', 'test.blogspam.net:8888',
        """URL of the BlogSpam service.""")

    skip_tests = ListOption('spam-filter', 'blogspam_skip_tests', '', doc=
        """Comma separated list of tests to skip.""")

    user_agent = 'Trac/%s | SpamFilter/%s'  % (
        TRAC_VERSION, get_distribution('TracSpamFilter').version
    )

    def __init__(self):
        self.external = True

    # IFilterStrategy implementation

    def test(self, req, author, content, ip):
        if not self._check_preconditions(req, author, content):
            return

        try:
            server = TimeoutServerProxy("http://"+self.api_url)
            res = server.testComment(self._getparams(req, author, content, ip))
            if res.startswith("SPAM:"):
                return -abs(self.karma_points), 'BlogSpam says content is spam (%s)' % res[5:]
        except Exception, v:
            self.log.warning('Checking with BlogSpam failed: %s', v)

    def train(self, req, author, content, ip, spam=True):
        if not self._check_preconditions(req, author, content):
            return

        try:
            params = self._getparams(req, author, content, ip)
            if spam:
                params['train'] = "spam"
            else:
                params['train'] = "ham"
            server = TimeoutServerProxy("http://"+self.api_url)
            res = server.classifyComment(params)
            self.log.debug('Classifying with BlogSpam succeeded: %s', res)
        except Exception, v:
            self.log.warning('Classifying with BlogSpam failed: %s', v)

    def getmethods(self):
        try:
            server=TimeoutServerProxy("http://"+self.api_url)
            return server.getPlugins();
        except Exception:
            return ""

    # Internal methods

    def _check_preconditions(self, req, author, content):
        if self.karma_points == 0:
            return False

        if len(content) == 0:
            return False

        if is_binary(content):
            self.log.warning('Content is binary, BlogSpam content check skipped')
            return False

        return True

    def _getparams(self, req, author, content, ip):
        # Split up author into name and email, if possible
        author = author.encode('utf-8')
        author_name, author_email = parseaddr(author)
        if not author_name and not author_email:
            author_name = author
        elif not author_name and author_email.find("@") < 1:
            author_name = author
            author_email = None
        params = {
            'ip':ip,
            'name':author_name,
            'comment':content.encode('utf-8'),
            'agent':req.get_header('User-Agent'),
            'site':req.base_url,
            'version':self.user_agent
        }
        if len(self.skip_tests):
            params['options'] = "exclude=%s" % ",exclude=".join(self.skip_tests)
        if author_email:
            params['email'] = author_email
        return params
